@@ -12,7 +12,8 @@ import (
 	"syscall"
 	"time"
 
-	"github.com/containers/libpod/pkg/cgroups"
+	"github.com/containers/podman/v3/pkg/cgroups"
+	"github.com/containers/podman/v3/pkg/signal"
 	"github.com/containers/storage/pkg/idtools"
 	ann "github.com/cri-o/cri-o/pkg/annotations"
 	json "github.com/json-iterator/go"
@@ -21,7 +22,6 @@ import (
 	"github.com/sirupsen/logrus"
 	"golang.org/x/sys/unix"
 	"k8s.io/apimachinery/pkg/fields"
-	pb "k8s.io/cri-api/pkg/apis/runtime/v1alpha2"
 	"k8s.io/kubernetes/pkg/kubelet/types"
 )
 
@@ -35,7 +35,6 @@ const (
 )
 
 var (
-	defaultStopSignal   = strconv.Itoa(defaultStopSignalInt)
 	ErrContainerStopped = errors.New("container is already stopped")
 	ErrNotFound         = errors.New("container process not found")
 	ErrNotInitialized   = errors.New("container PID not initialized")
@@ -64,7 +63,7 @@ type Container struct {
 	annotations        fields.Set
 	crioAnnotations    fields.Set
 	state              *ContainerState
-	metadata           *pb.ContainerMetadata
+	metadata           *Metadata
 	opLock             sync.RWMutex
 	spec               *specs.Spec
 	idMappings         *idtools.IDMappings
@@ -73,6 +72,22 @@ type Container struct {
 	stdinOnce          bool
 	created            bool
 	spoofed            bool
+	stopping           bool
+	stopTimeoutChan    chan time.Duration
+	stoppedChan        chan struct{}
+	stopStoppingChan   chan struct{}
+	stopLock           sync.Mutex
+}
+
+// Metadata holds all necessary information for building the container name.
+// The container runtime is encouraged to expose the metadata in its user
+// interface for better user experience.
+type Metadata struct {
+	// Name of the container.
+	Name string `json:"name,omitempty"`
+
+	// Attempt number of creating the container.
+	Attempt uint32 `json:"attempt,omitempty"`
 }
 
 // ContainerVolume is a bind mount for the container.
@@ -99,34 +114,37 @@ type ContainerState struct {
 }
 
 // NewContainer creates a container object.
-func NewContainer(id, name, bundlePath, logPath string, labels, crioAnnotations, annotations map[string]string, image, imageName, imageRef string, metadata *pb.ContainerMetadata, sandbox string, terminal, stdin, stdinOnce bool, runtimeHandler, dir string, created time.Time, stopSignal string) (*Container, error) {
+func NewContainer(id, name, bundlePath, logPath string, labels, crioAnnotations, annotations map[string]string, image, imageName, imageRef string, metadata *Metadata, sandbox string, terminal, stdin, stdinOnce bool, runtimeHandler, dir string, created time.Time, stopSignal string) (*Container, error) {
 	state := &ContainerState{}
 	state.Created = created
 	c := &Container{
-		id:              id,
-		name:            name,
-		bundlePath:      bundlePath,
-		logPath:         logPath,
-		labels:          labels,
-		sandbox:         sandbox,
-		terminal:        terminal,
-		stdin:           stdin,
-		stdinOnce:       stdinOnce,
-		runtimeHandler:  runtimeHandler,
-		metadata:        metadata,
-		annotations:     annotations,
-		crioAnnotations: crioAnnotations,
-		image:           image,
-		imageName:       imageName,
-		imageRef:        imageRef,
-		dir:             dir,
-		state:           state,
-		stopSignal:      stopSignal,
+		id:               id,
+		name:             name,
+		bundlePath:       bundlePath,
+		logPath:          logPath,
+		labels:           labels,
+		sandbox:          sandbox,
+		terminal:         terminal,
+		stdin:            stdin,
+		stdinOnce:        stdinOnce,
+		runtimeHandler:   runtimeHandler,
+		metadata:         metadata,
+		annotations:      annotations,
+		crioAnnotations:  crioAnnotations,
+		image:            image,
+		imageName:        imageName,
+		imageRef:         imageRef,
+		dir:              dir,
+		state:            state,
+		stopSignal:       stopSignal,
+		stopTimeoutChan:  make(chan time.Duration, 1),
+		stoppedChan:      make(chan struct{}, 1),
+		stopStoppingChan: make(chan struct{}, 1),
 	}
 	return c, nil
 }
 
-func NewSpoofedContainer(id, name string, labels map[string]string, created time.Time, dir string) *Container {
+func NewSpoofedContainer(id, name string, labels map[string]string, sandbox string, created time.Time, dir string) *Container {
 	state := &ContainerState{}
 	state.Created = created
 	state.Started = created
@@ -137,6 +155,7 @@ func NewSpoofedContainer(id, name string, labels map[string]string, created time
 		spoofed: true,
 		state:   state,
 		dir:     dir,
+		sandbox: sandbox,
 	}
 	c.annotations = map[string]string{
 		ann.SpoofedContainer: "true",
@@ -163,16 +182,9 @@ func (c *Container) ConmonCgroupfsPath() string {
 // GetStopSignal returns the container's own stop signal configured from the
 // image configuration or the default one.
 func (c *Container) GetStopSignal() string {
-	if c.stopSignal == "" {
-		return defaultStopSignal
-	}
-	signal := unix.SignalNum(strings.ToUpper(c.stopSignal))
-	if signal == 0 {
-		return defaultStopSignal
-	}
 	// return the stop signal in the form of its int converted to a string
 	// i.e stop signal 34 is returned as "34" to avoid back and forth conversion
-	return strconv.Itoa(int(signal))
+	return strconv.Itoa(int(c.StopSignal()))
 }
 
 // StopSignal returns the container's own stop signal configured from
@@ -182,11 +194,11 @@ func (c *Container) StopSignal() syscall.Signal {
 		return defaultStopSignalInt
 	}
 
-	signal := unix.SignalNum(strings.ToUpper(c.stopSignal))
-	if signal == 0 {
+	s, err := signal.ParseSignal(strings.ToUpper(c.stopSignal))
+	if err != nil {
 		return defaultStopSignalInt
 	}
-	return signal
+	return s
 }
 
 // FromDisk restores container's state from disk
@@ -337,7 +349,7 @@ func (c *Container) Dir() string {
 }
 
 // Metadata returns the metadata of the container.
-func (c *Container) Metadata() *pb.ContainerMetadata {
+func (c *Container) Metadata() *Metadata {
 	return c.metadata
 }
 
@@ -452,6 +464,9 @@ func (c *Container) pid() (int, error) {
 	if err := c.verifyPid(); err != nil {
 		return 0, err
 	}
+	if err := unix.Kill(c.state.InitPid, 0); err == unix.ESRCH {
+		return 0, errors.Wrapf(err, "check whether %d is running", c.state.InitPid)
+	}
 	return c.state.InitPid, nil
 }
 
@@ -533,3 +548,41 @@ func (c *Container) ShouldBeStopped() error {
 func (c *Container) Spoofed() bool {
 	return c.spoofed
 }
+
+// SetAsStopping marks a container as being stopped.
+// If a stop is currently happening, it also sends the new timeout
+// along the stopTimeoutChan, allowing the in-progress stop
+// to stop faster, or ignore the new stop timeout.
+// In this case, it also returns true, signifying the caller doesn't have to
+// Do any stop related cleanup, as the original caller (alreadyStopping=false)
+// will do said cleanup.
+func (c *Container) SetAsStopping(timeout int64) (alreadyStopping bool) {
+	// First, need to check if the container is already stopping
+	c.stopLock.Lock()
+	defer c.stopLock.Unlock()
+	if c.stopping {
+		// If so, we shouldn't wait forever on the opLock.
+		// This can cause issues where the container stop gets DOSed by a very long
+		// timeout, followed a shorter one coming in.
+		// Instead, interrupt the other stop with this new one.
+		select {
+		case c.stopTimeoutChan <- time.Duration(timeout) * time.Second:
+		case <-c.stoppedChan: // This case is to avoid waiting forever once another routine has finished.
+		case <-c.stopStoppingChan: // This case is to avoid deadlocking with SetAsNotStopping.
+		}
+		return true
+	}
+	// Regardless, set the container as actively stopping.
+	c.stopping = true
+	// And reset the stopStoppingChan
+	c.stopStoppingChan = make(chan struct{}, 1)
+	return false
+}
+
+// SetAsNotStopping unsets the stopping field indicating to new callers that the container
+// is no longer actively stopping.
+func (c *Container) SetAsNotStopping() {
+	c.stopLock.Lock()
+	c.stopping = false
+	c.stopLock.Unlock()
+}
