@@ -8,27 +8,30 @@ import (
 	"os"
 	"os/exec"
 	"path/filepath"
+	"regexp"
 	"strings"
-	"syscall"
 
 	"github.com/BurntSushi/toml"
 	conmonconfig "github.com/containers/conmon/runner/config"
 	"github.com/containers/image/v5/pkg/sysregistriesv2"
 	"github.com/containers/image/v5/types"
-	"github.com/containers/libpod/v2/pkg/hooks"
-	"github.com/containers/libpod/v2/pkg/rootless"
+	"github.com/containers/podman/v3/pkg/hooks"
+	"github.com/containers/podman/v3/pkg/rootless"
 	"github.com/containers/storage"
 	"github.com/cri-o/cri-o/internal/config/apparmor"
 	"github.com/cri-o/cri-o/internal/config/capabilities"
 	"github.com/cri-o/cri-o/internal/config/cgmgr"
+	"github.com/cri-o/cri-o/internal/config/cnimgr"
 	"github.com/cri-o/cri-o/internal/config/conmonmgr"
 	"github.com/cri-o/cri-o/internal/config/device"
 	"github.com/cri-o/cri-o/internal/config/node"
 	"github.com/cri-o/cri-o/internal/config/nsmgr"
 	"github.com/cri-o/cri-o/internal/config/seccomp"
 	"github.com/cri-o/cri-o/internal/config/ulimits"
+	"github.com/cri-o/cri-o/pkg/annotations"
 	"github.com/cri-o/cri-o/server/useragent"
 	"github.com/cri-o/cri-o/utils"
+	"github.com/cri-o/cri-o/utils/cmdrunner"
 	"github.com/cri-o/ocicni/pkg/ocicni"
 	selinux "github.com/opencontainers/selinux/go-selinux"
 	"github.com/pkg/errors"
@@ -39,13 +42,16 @@ import (
 
 // Defaults if none are specified
 const (
-	defaultRuntime        = "runc"
-	DefaultRuntimeType    = "oci"
-	DefaultRuntimeRoot    = "/run/runc"
-	defaultGRPCMaxMsgSize = 16 * 1024 * 1024
-	OCIBufSize            = 8192
-	RuntimeTypeVM         = "vm"
-	defaultCtrStopTimeout = 30 // seconds
+	defaultRuntime             = "runc"
+	DefaultRuntimeType         = "oci"
+	DefaultRuntimeRoot         = "/run/runc"
+	defaultGRPCMaxMsgSize      = 16 * 1024 * 1024
+	OCIBufSize                 = 8192
+	RuntimeTypeVM              = "vm"
+	defaultCtrStopTimeout      = 30 // seconds
+	defaultNamespacesDir       = "/var/run"
+	RuntimeTypeVMBinaryPattern = "containerd-shim-([a-zA-Z0-9\\-\\+])+-v2"
+	tasksetBinary              = "taskset"
 )
 
 // Config represents the entire set of configuration values that can be set for
@@ -94,6 +100,8 @@ const (
 	ImageVolumesIgnore ImageVolumesType = "ignore"
 	// ImageVolumesBind option is for using bind mounted volumes
 	ImageVolumesBind ImageVolumesType = "bind"
+	// DefaultPauseImage is default pause image
+	DefaultPauseImage string = "k8s.gcr.io/pause:3.5"
 )
 
 const (
@@ -104,10 +112,6 @@ const (
 	// DefaultLogSizeMax is the default value for the maximum log size
 	// allowed for a container. Negative values mean that no limit is imposed.
 	DefaultLogSizeMax = -1
-
-	// DefaultLogToJournald is the default value for whether conmon should
-	// log to journald in addition to kubernetes log file.
-	DefaultLogToJournald = false
 )
 
 const (
@@ -147,6 +151,14 @@ type RootConfig struct {
 	// VersionFilePersist is the location CRI-O will lay down the version file
 	// that checks whether we've upgraded
 	VersionFilePersist string `toml:"version_file_persist"`
+
+	// CleanShutdownFile is the location CRI-O will lay down the clean shutdown file
+	// that checks whether we've had time to sync before shutting down
+	CleanShutdownFile string `toml:"clean_shutdown_file"`
+
+	// InternalWipe is whether CRI-O should wipe containers and images after a reboot when the server starts.
+	// If set to false, one must use the external command `crio wipe` to wipe the containers and images in these situations.
+	InternalWipe bool `toml:"internal_wipe"`
 }
 
 // RuntimeHandler represents each item of the "crio.runtime.runtimes" TOML
@@ -163,7 +175,12 @@ type RuntimeHandler struct {
 	// "io.kubernetes.cri-o.userns-mode" for configuring a user namespace for the pod.
 	// "io.kubernetes.cri-o.Devices" for configuring devices for the pod.
 	// "io.kubernetes.cri-o.ShmSize" for configuring the size of /dev/shm.
+	// "io.kubernetes.cri-o.UnifiedCgroup.$CTR_NAME" for configuring the cgroup v2 unified block for a container.
+	// "io.containers.trace-syscall" for tracing syscalls via the OCI seccomp BPF hook.
 	AllowedAnnotations []string `toml:"allowed_annotations,omitempty"`
+
+	// DisallowedAnnotations is the slice of experimental annotations that are not allowed for this handler.
+	DisallowedAnnotations []string
 }
 
 // Multiple runtime Handlers in a map
@@ -185,13 +202,7 @@ type RuntimeConfig struct {
 	// to the kubernetes log file
 	LogToJournald bool `toml:"log_to_journald"`
 
-	// ManageNSLifecycle determines whether we pin and remove namespaces
-	// and manage their lifecycle
-	// This option is being deprecated
-	ManageNSLifecycle bool `toml:"manage_ns_lifecycle"`
-
 	// DropInfraCtr determines whether the infra container is dropped when appropriate.
-	// Requires ManageNSLifecycle to be true.
 	DropInfraCtr bool `toml:"drop_infra_ctr"`
 
 	// ReadOnly run all pods/containers in read-only mode.
@@ -280,7 +291,7 @@ type RuntimeConfig struct {
 	GIDMappings string `toml:"gid_mappings"`
 
 	// LogLevel determines the verbosity of the logs based on the level it is set to.
-	// Options are fatal, panic, error (default), warn, info, and debug.
+	// Options are fatal, panic, error (default), warn, info, debug, and trace.
 	LogLevel string `toml:"log_level"`
 
 	// LogFilter specifies a regular expression to filter the log messages
@@ -300,6 +311,10 @@ type RuntimeConfig struct {
 	// the level of trust of the workload.
 	Runtimes Runtimes `toml:"runtimes"`
 
+	// Workloads defines a list of workloads types that are have grouped settings
+	// that will be applied to containers.
+	Workloads Workloads `toml:"workloads"`
+
 	// PidsLimit is the number of processes each container is restricted to
 	// by the cgroup process number controller.
 	PidsLimit int64 `toml:"pids_limit"`
@@ -320,6 +335,10 @@ type RuntimeConfig struct {
 	// InfraCtrCPUSet is the CPUs set that will be used to run infra containers
 	InfraCtrCPUSet string `toml:"infra_ctr_cpuset"`
 
+	// AbsentMountSourcesToReject is a list of paths that, when absent from the host,
+	// will cause a container creation to fail (as opposed to the current behavior of creating a directory).
+	AbsentMountSourcesToReject []string `toml:"absent_mount_sources_to_reject"`
+
 	// seccompConfig is the internal seccomp configuration
 	seccompConfig *seccomp.Config
 
@@ -337,6 +356,9 @@ type RuntimeConfig struct {
 
 	// conmonManager is the internal ConmonManager configuration
 	conmonManager *conmonmgr.ConmonManager
+
+	// namespaceManager is the internal NamespaceManager configuration
+	namespaceManager *nsmgr.NamespaceManager
 }
 
 // ImageConfig represents the "crio.image" TOML config table.
@@ -389,8 +411,8 @@ type NetworkConfig struct {
 	// PluginDirs is where CNI plugin binaries are stored.
 	PluginDirs []string `toml:"plugin_dirs"`
 
-	// cniPlugin is the internal OCI CNI plugin
-	cniPlugin ocicni.CNIPlugin
+	// cniManager manages the internal ocicni plugin
+	cniManager *cnimgr.CNIManager
 }
 
 // APIConfig represents the "crio.api" TOML config table.
@@ -480,10 +502,29 @@ func (t *tomlConfig) fromConfig(c *Config) {
 	t.Crio.Metrics.MetricsConfig = c.MetricsConfig
 }
 
-// UpdateFromFile populates the Config from the TOML-encoded file at the given path.
+// UpdateFromFile populates the Config from the TOML-encoded file at the given
+// path and "remembers" that we should reload this file's contents when we
+// receive a SIGHUP.
 // Returns errors encountered when reading or parsing the files, or nil
 // otherwise.
 func (c *Config) UpdateFromFile(path string) error {
+	if err := c.UpdateFromDropInFile(path); err != nil {
+		return err
+	}
+	c.singleConfigPath = path
+	return nil
+}
+
+// UpdateFromDropInFile populates the Config from the TOML-encoded file at the
+// given path.  The file may be the main configuration file, or it can be one
+// of the drop-in files which are used to supplement it.
+// Returns errors encountered when reading or parsing the files, or nil
+// otherwise.
+func (c *Config) UpdateFromDropInFile(path string) error {
+	// keeps the storage options from storage.conf and merge it to crio config
+	var storageOpts []string
+	storageOpts = append(storageOpts, c.StorageOptions...)
+
 	data, err := ioutil.ReadFile(path)
 	if err != nil {
 		return err
@@ -503,20 +544,43 @@ func (c *Config) UpdateFromFile(path string) error {
 		delete(c.Runtimes, defaultRuntime)
 	}
 
+	storageOpts = append(storageOpts, t.Crio.RootConfig.StorageOptions...)
+	storageOpts = removeDupStorageOpts(storageOpts)
+	t.Crio.RootConfig.StorageOptions = storageOpts
+
+	// Registries are deprecated in cri-o.conf and turned into a NOP.
+	// Users should use registries.conf instead, so let's log it.
 	if len(t.Crio.Image.Registries) > 0 {
-		logrus.Warnf("The 'registries' option in crio.conf(5) (referenced in %q) has been deprecated and will be removed with CRI-O 1.21.", path)
-		logrus.Warn("Please refer to containers-registries.conf(5) for configuring unqualified-search registries.")
+		t.Crio.Image.Registries = nil
+		logrus.Warnf("Support for the 'registries' option has been dropped but it is referenced in %q.  Please use containers-registries.conf(5) for configuring unqualified-search registries instead.", path)
 	}
 
 	t.toConfig(c)
-	c.singleConfigPath = path
 	return nil
 }
 
+// removeDupStorageOpts removes duplicated storage option from the list
+// keeps the last appearance
+func removeDupStorageOpts(storageOpts []string) []string {
+	var resOpts []string
+	opts := make(map[string]bool)
+	for i := len(storageOpts) - 1; i >= 0; i-- {
+		if ok := opts[storageOpts[i]]; ok {
+			continue
+		}
+		opts[storageOpts[i]] = true
+		resOpts = append(resOpts, storageOpts[i])
+	}
+	for i, j := 0, len(resOpts)-1; i < j; i, j = i+1, j-1 {
+		resOpts[i], resOpts[j] = resOpts[j], resOpts[i]
+	}
+	return resOpts
+}
+
 // UpdateFromPath recursively iterates the provided path and updates the
 // configuration for it
 func (c *Config) UpdateFromPath(path string) error {
-	if _, err := os.Stat(path); os.IsNotExist(err) {
+	if _, err := os.Stat(path); err != nil && os.IsNotExist(err) {
 		return nil
 	}
 	if err := filepath.Walk(path,
@@ -527,7 +591,7 @@ func (c *Config) UpdateFromPath(path string) error {
 			if info.IsDir() {
 				return nil
 			}
-			return c.UpdateFromFile(p)
+			return c.UpdateFromDropInFile(p)
 		}); err != nil {
 		return err
 	}
@@ -582,6 +646,7 @@ func DefaultConfig() (*Config, error) {
 			LogDir:             "/var/log/crio/pods",
 			VersionFile:        CrioVersionPathTmp,
 			VersionFilePersist: CrioVersionPathPersist,
+			CleanShutdownFile:  CrioCleanShutdownFile,
 		},
 		APIConfig: APIConfig{
 			Listen:             CrioSocketPath,
@@ -597,6 +662,9 @@ func DefaultConfig() (*Config, error) {
 				defaultRuntime: {
 					RuntimeType: DefaultRuntimeType,
 					RuntimeRoot: DefaultRuntimeRoot,
+					AllowedAnnotations: []string{
+						annotations.OCISeccompBPFHookAnnotation,
+					},
 				},
 			},
 			ConmonEnv: []string{
@@ -615,17 +683,17 @@ func DefaultConfig() (*Config, error) {
 			DefaultCapabilities:      capabilities.Default(),
 			LogLevel:                 "info",
 			HooksDir:                 []string{hooks.DefaultDir},
-			NamespacesDir:            "/var/run",
-			ManageNSLifecycle:        true,
+			NamespacesDir:            defaultNamespacesDir,
 			seccompConfig:            seccomp.New(),
 			apparmorConfig:           apparmor.New(),
 			ulimitsConfig:            ulimits.New(),
 			cgroupManager:            cgroupManager,
 			deviceConfig:             device.New(),
+			namespaceManager:         nsmgr.New(defaultNamespacesDir, ""),
 		},
 		ImageConfig: ImageConfig{
 			DefaultTransport: "docker://",
-			PauseImage:       "k8s.gcr.io/pause:3.2",
+			PauseImage:       DefaultPauseImage,
 			PauseCommand:     "/pause",
 			ImageVolumes:     ImageVolumesMkdir,
 		},
@@ -750,6 +818,10 @@ func (c *RootConfig) Validate(onExecution bool) error {
 	return nil
 }
 
+func (c *RootConfig) CleanShutdownSupportedFileName() string {
+	return c.CleanShutdownFile + ".supported"
+}
+
 // Validate is the main entry point for runtime configuration validation
 // The parameter `onExecution` specifies if the validation should include
 // execution checks. It returns an `error` on validation failure, otherwise
@@ -788,14 +860,6 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 		return errors.New("conmon cgroup should be 'pod' or a systemd slice")
 	}
 
-	if !c.ManageNSLifecycle {
-		logrus.Infof("The manage-ns-lifecycle option is being deprecated, and will be unconditionally true in the future")
-	}
-
-	if c.DropInfraCtr && !c.ManageNSLifecycle {
-		return errors.New("cannot drop infra without ManageNSLifecycle")
-	}
-
 	if c.LogSizeMax >= 0 && c.LogSizeMax < OCIBufSize {
 		return fmt.Errorf("log size max should be negative or >= %d", OCIBufSize)
 	}
@@ -817,9 +881,20 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 	}
 
 	if c.InfraCtrCPUSet != "" {
-		if _, err := cpuset.Parse(c.InfraCtrCPUSet); err != nil {
+		set, err := cpuset.Parse(c.InfraCtrCPUSet)
+		if err != nil {
 			return errors.Wrap(err, "invalid infra_ctr_cpuset")
 		}
+
+		executable, err := exec.LookPath(tasksetBinary)
+		if err != nil {
+			return errors.Wrapf(err, "%q not found in $PATH", tasksetBinary)
+		}
+		cmdrunner.PrependCommandsWith(executable, "--cpu-list", set.String())
+	}
+
+	if err := c.Workloads.Validate(); err != nil {
+		return errors.Wrap(err, "workloads validation")
 	}
 
 	// check for validation on execution
@@ -865,30 +940,9 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 			return errors.Wrap(err, "pinns validation")
 		}
 
-		if err := os.MkdirAll(c.NamespacesDir, 0o755); err != nil {
-			return errors.Wrap(err, "invalid namespaces_dir")
-		}
-
-		for _, ns := range nsmgr.SupportedNamespacesForPinning() {
-			nsDir := filepath.Join(c.NamespacesDir, string(ns)+"ns")
-			if err := utils.IsDirectory(nsDir); err != nil {
-				// The file is not a directory, but exists.
-				// We should remove it.
-				if errors.Is(err, syscall.ENOTDIR) {
-					if err := os.Remove(nsDir); err != nil {
-						return errors.Wrapf(err, "remove file to create namespaces sub-dir")
-					}
-					logrus.Infof("Removed file %s to create directory in that path.", nsDir)
-				} else if !os.IsNotExist(err) {
-					// if it's neither an error because the file exists
-					// nor an error because it does not exist, it is
-					// some other disk error.
-					return errors.Wrapf(err, "checking whether namespaces sub-dir exists")
-				}
-				if err := os.MkdirAll(nsDir, 0o755); err != nil {
-					return errors.Wrap(err, "invalid namespaces sub-dir")
-				}
-			}
+		c.namespaceManager = nsmgr.New(c.NamespacesDir, c.PinnsPath)
+		if err := c.namespaceManager.Initialize(); err != nil {
+			return errors.Wrapf(err, "initialize nsmgr")
 		}
 
 		if c.SeccompUseDefaultWhenEmpty {
@@ -918,12 +972,24 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 
 // ValidateRuntimes checks every runtime if its members are valid
 func (c *RuntimeConfig) ValidateRuntimes() error {
+	var failedValidation []string
+
 	// Validate if runtime_path does exist for each runtime
 	for name, handler := range c.Runtimes {
 		if err := handler.Validate(name); err != nil {
-			return err
+			if c.DefaultRuntime == name {
+				return err
+			}
+
+			logrus.Warnf("'%s is being ignored due to: %q", name, err)
+			failedValidation = append(failedValidation, name)
 		}
 	}
+
+	for _, invalidHandlerName := range failedValidation {
+		delete(c.Runtimes, invalidHandlerName)
+	}
+
 	return nil
 }
 
@@ -967,6 +1033,11 @@ func (c *RuntimeConfig) CgroupManager() cgmgr.CgroupManager {
 	return c.cgroupManager
 }
 
+// NamespaceManager returns the NamespaceManager configuration
+func (c *RuntimeConfig) NamespaceManager() *nsmgr.NamespaceManager {
+	return c.namespaceManager
+}
+
 // Ulimits returns the Ulimits configuration
 func (c *RuntimeConfig) Ulimits() []ulimits.Ulimit {
 	return c.ulimitsConfig.Ulimits()
@@ -1030,13 +1101,13 @@ func (c *NetworkConfig) Validate(onExecution bool) error {
 		}
 
 		// Init CNI plugin
-		cniPlugin, err := ocicni.InitCNI(
+		cniManager, err := cnimgr.New(
 			c.CNIDefaultNetwork, c.NetworkDir, c.PluginDirs...,
 		)
 		if err != nil {
 			return errors.Wrap(err, "initialize CNI plugin")
 		}
-		c.cniPlugin = cniPlugin
+		c.cniManager = cniManager
 	}
 
 	return nil
@@ -1047,9 +1118,27 @@ func (r *RuntimeHandler) Validate(name string) error {
 	if err := r.ValidateRuntimePath(name); err != nil {
 		return err
 	}
+	if err := r.ValidateRuntimeAllowedAnnotations(); err != nil {
+		return err
+	}
 	return r.ValidateRuntimeType(name)
 }
 
+func (r *RuntimeHandler) ValidateRuntimeVMBinaryPattern() bool {
+	if r.RuntimeType != RuntimeTypeVM {
+		return true
+	}
+
+	binaryName := filepath.Base(r.RuntimePath)
+
+	matched, err := regexp.MatchString(RuntimeTypeVMBinaryPattern, binaryName)
+	if err != nil {
+		return false
+	}
+
+	return matched
+}
+
 // ValidateRuntimePath checks if the `RuntimePath` is either set or available
 // within the $PATH environment. The method fails on any `RuntimePath` lookup
 // error.
@@ -1061,12 +1150,20 @@ func (r *RuntimeHandler) ValidateRuntimePath(name string) error {
 		}
 		r.RuntimePath = executable
 		logrus.Debugf("using runtime executable from $PATH %q", executable)
-	} else if _, err := os.Stat(r.RuntimePath); os.IsNotExist(err) {
+	} else if _, err := os.Stat(r.RuntimePath); err != nil && os.IsNotExist(err) {
 		return fmt.Errorf("invalid runtime_path for runtime '%s': %q",
 			name, err)
 	}
-	logrus.Debugf("found valid runtime %q for runtime_path %q",
-		name, r.RuntimePath)
+
+	ok := r.ValidateRuntimeVMBinaryPattern()
+	if !ok {
+		return fmt.Errorf("invalid runtime_path for runtime '%s': containerd binary naming pattern is not followed",
+			name)
+	}
+
+	logrus.Debugf(
+		"Found valid runtime %q for runtime_path %q", name, r.RuntimePath,
+	)
 	return nil
 }
 
@@ -1079,12 +1176,42 @@ func (r *RuntimeHandler) ValidateRuntimeType(name string) error {
 	return nil
 }
 
-func (c *Config) SetLocations(singleConfigPath, dropInConfigDir string) {
-	c.singleConfigPath = singleConfigPath
-	c.dropInConfigDir = dropInConfigDir
+func (r *RuntimeHandler) ValidateRuntimeAllowedAnnotations() error {
+	disallowedAnnotations := make(map[string]struct{})
+	for _, ann := range annotations.AllAllowedAnnotations {
+		disallowedAnnotations[ann] = struct{}{}
+	}
+	for _, allowed := range r.AllowedAnnotations {
+		if _, ok := disallowedAnnotations[allowed]; !ok {
+			return errors.Errorf("invalid allowed_annotation: %s", allowed)
+		}
+		delete(disallowedAnnotations, allowed)
+	}
+	for ann := range disallowedAnnotations {
+		r.DisallowedAnnotations = append(r.DisallowedAnnotations, ann)
+	}
+	logrus.Debugf(
+		"Allowed annotations for runtime: %v", r.AllowedAnnotations,
+	)
+	return nil
 }
 
 // CNIPlugin returns the network configuration CNI plugin
 func (c *NetworkConfig) CNIPlugin() ocicni.CNIPlugin {
-	return c.cniPlugin
+	return c.cniManager.Plugin()
+}
+
+// CNIPluginReadyOrError returns whether the cni plugin is ready
+func (c *NetworkConfig) CNIPluginReadyOrError() error {
+	return c.cniManager.ReadyOrError()
+}
+
+// CNIPluginAddWatcher returns the network configuration CNI plugin
+func (c *NetworkConfig) CNIPluginAddWatcher() chan struct{} {
+	return c.cniManager.AddWatcher()
+}
+
+// SetSingleConfigPath set single config path for config
+func (c *Config) SetSingleConfigPath(singleConfigPath string) {
+	c.singleConfigPath = singleConfigPath
 }
